"""Chromobius (Development Version): an implementation of the mobius color code decoder."""
# (This a stubs file describing the classes and methods in stim.)
from __future__ import annotations
from typing import overload, TYPE_CHECKING, Any, Iterable
if TYPE_CHECKING:
    import io
    import pathlib
    import numpy as np
    import stim
    import sinter
    import chromobius
__version__: str
class CompiledDecoder:
    """A chromobius decoder ready to predict observable flips from detection events.

    Example:
        >>> import stim
        >>> import chromobius

        >>> dem = stim.Circuit('''
        ...     X_ERROR(0.1) 0 1 2 3 4 5 6 7
        ...     MPP Z0*Z1*Z2 Z1*Z2*Z3 Z2*Z3*Z4 Z3*Z4*Z5
        ...     DETECTOR(0, 0, 0, 1) rec[-4]
        ...     DETECTOR(1, 0, 0, 2) rec[-3]
        ...     DETECTOR(2, 0, 0, 0) rec[-2]
        ...     DETECTOR(3, 0, 0, 1) rec[-1]
        ...     M 0
        ...     OBSERVABLE_INCLUDE(0) rec[-1]
        ... ''').detector_error_model()

        >>> decoder = chromobius.CompiledDecoder.from_dem(dem)
    """
    @staticmethod
    def from_dem(
        dem: stim.DetectorErrorModel,
    ) -> chromobius.CompiledDecoder:
        """Compiles a decoder for a stim detector error model.

        Args:
            dem: A stim detector error model. The detector error model must satisfy:
                1. Basis+Color annotations. Every detector that appears in an error
                    must specify coordinate data including a fourth coordinate. The 4th
                    coordinate indicates the basis and color of the detector with the
                    convention:
                        0 = Red X
                        1 = Green X
                        2 = Blue X
                        3 = Red Z
                        4 = Green Z
                        5 = Blue Z
                        -1 = Ignore this Detector
                2. Rainbow triplets. Bulk errors with three symptoms in one basis should
                    have one symptom of each color. Errors with three symptoms that
                    repeat a color will cause an exception unless they can be decomposed
                    into other basic errors.
                3. Movable excitations. It needs to be possible to combine bulk errors
                    to form simpler errors with one or two symptoms that can be used to
                    move or destroy excitations. If bulk errors don't have this
                    property, decoding will fail when attempting to lift a solution
                    from the matcher requires dragging an excitation along a path but
                    there's no way to move the excitation along that path.
                4. Matchable-avoids-color. In parts of the dem that correspond to a
                    matchable code, at least one of the colors must be avoided.
                    Otherwise the matcher may be given a problem that can be solved
                    locally, but when lifting it needs to be solved non-locally.

        Returns:
            A decoder object that can be used to predict observable flips from
            detection event samples.

        Example:
            >>> import stim
            >>> import chromobius

            >>> dem = stim.Circuit('''
            ...     X_ERROR(0.1) 0 1 2 3 4 5 6 7
            ...     MPP Z0*Z1*Z2 Z1*Z2*Z3 Z2*Z3*Z4 Z3*Z4*Z5
            ...     DETECTOR(0, 0, 0, 1) rec[-4]
            ...     DETECTOR(1, 0, 0, 2) rec[-3]
            ...     DETECTOR(2, 0, 0, 0) rec[-2]
            ...     DETECTOR(3, 0, 0, 1) rec[-1]
            ...     M 0
            ...     OBSERVABLE_INCLUDE(0) rec[-1]
            ... ''').detector_error_model()

            >>> decoder = chromobius.CompiledDecoder.from_dem(dem)
        """
    @staticmethod
    def predict_obs_flips_from_dets_bit_packed(
        dets: np.ndarray,
    ) -> np.ndarray:
        """Predicts observable flips from detection events.

        Args:
            dets: A bit packed numpy array of detection event data. The array can either
                be 1-dimensional (a single shot to decode) or 2-dimensional (multiple
                shots to decode, with the first axis being the shot axis and the second
                axis being the detection event byte axis).

                The array's dtype must be np.uint8. If you have an array of dtype
                np.bool_, you have data that's not bit packed. You can pack it by
                using `np.packbits(array, bitorder='little')`. But ideally you
                should attempt to never have unpacked data in the first place,
                since it's 8x larger which can be a large performance loss. For
                example, stim's sampler methods all have a `bit_packed=True` argument
                that cause them to return bit packed data.

        Returns:
            A bit packed numpy array of observable flip data. The array will have
            the same number of dimensions as the dets argument.

            If dets is a 1D array, then the result has:
                shape = (math.ceil(num_obs / 8),)
                dtype = np.uint8
            If dets is a 2D array, then the result has:
                shape = (dets.shape[0], math.ceil(num_obs / 8),)
                dtype = np.uint8

            To determine if the observable with index k was flipped in shot s, compute:
                `bool((result[s, k // 8] >> (k % 8)) & 1)`

        Example:
            >>> import stim
            >>> import chromobius
            >>> import numpy as np

            >>> repetition_color_code = stim.Circuit('''
            ...     # Apply noise.
            ...     X_ERROR(0.1) 0 1 2 3 4 5 6 7
            ...     # Measure three-body stabilizers to catch errors.
            ...     MPP Z0*Z1*Z2 Z1*Z2*Z3 Z2*Z3*Z4 Z3*Z4*Z5 Z4*Z5*Z6 Z5*Z6*Z7
            ...
            ...     # Annotate detectors, with a coloring in the 4th coordinate.
            ...     DETECTOR(0, 0, 0, 2) rec[-6]
            ...     DETECTOR(1, 0, 0, 0) rec[-5]
            ...     DETECTOR(2, 0, 0, 1) rec[-4]
            ...     DETECTOR(3, 0, 0, 2) rec[-3]
            ...     DETECTOR(4, 0, 0, 0) rec[-2]
            ...     DETECTOR(5, 0, 0, 1) rec[-1]
            ...
            ...     # Check on the message.
            ...     M 0
            ...     OBSERVABLE_INCLUDE(0) rec[-1]
            ... ''')

            >>> # Sample the circuit.
            >>> shots = 4096
            >>> sampler = repetition_color_code.compile_detector_sampler()
            >>> dets, actual_obs_flips = sampler.sample(
            ...     shots=shots,
            ...     separate_observables=True,
            ...     bit_packed=True,
            ... )

            >>> # Decode with Chromobius.
            >>> dem = repetition_color_code.detector_error_model()
            >>> decoder = chromobius.compile_decoder_for_dem(dem)
            >>> predicted_flips = decoder.predict_obs_flips_from_dets_bit_packed(dets)

            >>> # Count mistakes.
            >>> differences = np.any(predicted_flips != actual_obs_flips, axis=1)
            >>> mistakes = np.count_nonzero(differences)
            >>> assert mistakes < shots / 5
        """
    @staticmethod
    def predict_weighted_obs_flips_from_dets_bit_packed(
        dets: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Predicts observable flips and weights from detection events.

        The returned weight comes directly from the underlying call to pymatching, not
        accounting for the lifting process.

        Args:
            dets: A bit packed numpy array of detection event data. The array can either
                be 1-dimensional (a single shot to decode) or 2-dimensional (multiple
                shots to decode, with the first axis being the shot axis and the second
                axis being the detection event byte axis).

                The array's dtype must be np.uint8. If you have an array of dtype
                np.bool_, you have data that's not bit packed. You can pack it by
                using `np.packbits(array, bitorder='little')`. But ideally you
                should attempt to never have unpacked data in the first place,
                since it's 8x larger which can be a large performance loss. For
                example, stim's sampler methods all have a `bit_packed=True` argument
                that cause them to return bit packed data.

        Returns:
            A tuple (obs, weights).
            Obs is a bit packed numpy array of observable flip data.
            Weights is a numpy array (or scalar) of floats.

            If dets is a 1D array, then the result has:
                obs.shape = (math.ceil(num_obs / 8),)
                obs.dtype = np.uint8
                weights.shape = ()
                weights.dtype = np.float32
            If dets is a 2D array, then the result has:
                shape = (dets.shape[0], math.ceil(num_obs / 8),)
                dtype = np.uint8
                weights.shape = (dets.shape[0],)
                weights.dtype = np.float32

            To determine if the observable with index k was flipped in shot s, compute:
                `bool((obs[s, k // 8] >> (k % 8)) & 1)`

        Example:
            >>> import stim
            >>> import chromobius
            >>> import numpy as np

            >>> repetition_color_code = stim.Circuit('''
            ...     # Apply noise.
            ...     X_ERROR(0.1) 0 1 2 3 4 5 6 7
            ...     # Measure three-body stabilizers to catch errors.
            ...     MPP Z0*Z1*Z2 Z1*Z2*Z3 Z2*Z3*Z4 Z3*Z4*Z5 Z4*Z5*Z6 Z5*Z6*Z7
            ...
            ...     # Annotate detectors, with a coloring in the 4th coordinate.
            ...     DETECTOR(0, 0, 0, 2) rec[-6]
            ...     DETECTOR(1, 0, 0, 0) rec[-5]
            ...     DETECTOR(2, 0, 0, 1) rec[-4]
            ...     DETECTOR(3, 0, 0, 2) rec[-3]
            ...     DETECTOR(4, 0, 0, 0) rec[-2]
            ...     DETECTOR(5, 0, 0, 1) rec[-1]
            ...
            ...     # Check on the message.
            ...     M 0
            ...     OBSERVABLE_INCLUDE(0) rec[-1]
            ... ''')

            >>> # Sample the circuit.
            >>> shots = 4096
            >>> sampler = repetition_color_code.compile_detector_sampler()
            >>> dets, actual_obs_flips = sampler.sample(
            ...     shots=shots,
            ...     separate_observables=True,
            ...     bit_packed=True,
            ... )

            >>> # Decode with Chromobius.
            >>> dem = repetition_color_code.detector_error_model()
            >>> decoder = chromobius.compile_decoder_for_dem(dem)
            >>> result = decoder.predict_weighted_obs_flips_from_dets_bit_packed(dets)
            >>> pred, weights = result
        """
def compile_decoder_for_dem(
    dem: stim.DetectorErrorModel,
) -> chromobius.CompiledDecoder:
    """Compiles a decoder for a stim detector error model.

    Args:
        dem: A stim detector error model. The detector error model must satisfy:
            1. Basis+Color annotations. Every detector that appears in an error
                must specify coordinate data including a fourth coordinate. The 4th
                coordinate indicates the basis and color of the detector with the
                convention:
                    0 = Red X
                    1 = Green X
                    2 = Blue X
                    3 = Red Z
                    4 = Green Z
                    5 = Blue Z
                    -1 = Ignore this Detector
            2. Rainbow triplets. Bulk errors with three symptoms in one basis should
                have one symptom of each color. Errors with three symptoms that
                repeat a color will cause an exception unless they can be decomposed
                into other basic errors.
            3. Movable excitations. It needs to be possible to combine bulk errors
                to form simpler errors with one or two symptoms that can be used to
                move or destroy excitations. If bulk errors don't have this
                property, decoding will fail when attempting to lift a solution
                from the matcher requires dragging an excitation along a path but
                there's no way to move the excitation along that path.
            4. Matchable-avoids-color. In parts of the dem that correspond to a
                matchable code, at least one of the colors must be avoided.
                Otherwise the matcher may be given a problem that can be solved
                locally, but when lifting it needs to be solved non-locally.

    Returns:
        A decoder object that can be used to predict observable flips from
        detection event samples.

    Example:
        >>> import stim
        >>> import chromobius

        >>> dem = stim.Circuit('''
        ...     X_ERROR(0.1) 0 1 2 3 4 5 6 7
        ...     MPP Z0*Z1*Z2 Z1*Z2*Z3 Z2*Z3*Z4 Z3*Z4*Z5
        ...     DETECTOR(0, 0, 0, 1) rec[-4]
        ...     DETECTOR(1, 0, 0, 2) rec[-3]
        ...     DETECTOR(2, 0, 0, 0) rec[-2]
        ...     DETECTOR(3, 0, 0, 1) rec[-1]
        ...     M 0
        ...     OBSERVABLE_INCLUDE(0) rec[-1]
        ... ''').detector_error_model()

        >>> decoder = chromobius.compile_decoder_for_dem(dem)
    """
def sinter_decoders() -> dict[str, sinter.Decoder]:
    """A dictionary describing chromobius to sinter.

    Giving the result of this function to the `custom_decoders` argument of
    `sinter.collect` will tell sinter about the decoder 'chromobius'. On the
    command line, the equivalent argument is
    `--custom_decoders 'chromobius:sinter_decoders'`.

    Returns:
        The dict `{'chromobius': <an object compatible with sinter.Decoder>}`.
    """



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import os
import pathlib
import re
import shutil
import subprocess
import sys

import setuptools
import setuptools.command.build_ext


CC_FILES = glob.glob("src/**/*.cc", recursive=True)
H_FILES = glob.glob("src/**/*.h", recursive=True) + glob.glob("src/**/*.inl", recursive=True)
CMAKE_FILES = ['CMakeLists.txt', *glob.glob("file_lists/*", recursive=True)]
RELEVANT_SOURCE_FILES = sorted(CMAKE_FILES + CC_FILES + H_FILES)


class CMakeExtension(setuptools.Extension):
    def __init__(self, name, sourcedir, sources):
        setuptools.Extension.__init__(self, name, sources=sources)
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(setuptools.command.build_ext.build_ext):
    def build_extension(self, ext):
        import ninja

        build_temp = os.path.join(self.build_temp, ext.name)
        os.makedirs(build_temp, exist_ok=True)
        build_env = {
            **os.environ,
            'CMAKE_LIBRARY_OUTPUT_DIRECTORY': '',
            'CMAKE_FORCE_PYBIND_CHROMOBIUS': '1',
        }

        osx_cmake_flags = []
        if sys.platform.startswith("darwin"):
            # Cross-compile support for macOS - respect ARCHFLAGS if set
            archs = re.findall(r"-arch (\S+)", os.environ.get("ARCHFLAGS", ""))
            if archs:
                osx_cmake_flags = ["-DCMAKE_OSX_ARCHITECTURES={}".format(";".join(archs))]
            else:
                import platform
                arch = platform.machine()
                if arch:
                    osx_cmake_flags = [f"-DCMAKE_OSX_ARCHITECTURES={arch}"]

        subprocess.check_call([
            "cmake",
            ext.sourcedir,
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={pathlib.Path(self.get_ext_fullpath(ext.name)).parent.absolute()}",
            f"-DPYTHON_EXECUTABLE={sys.executable}",
            f"-DCHROMOBIUS_VERSION_INFO={__version__}",
            *osx_cmake_flags,
            *[
                env_arg_item
                for env_arg_item in os.environ.get("CMAKE_ARGS", "").split(" ")
                if env_arg_item
            ],
            "-GNinja",
            f"-DCMAKE_MAKE_PROGRAM:FILEPATH={os.path.join(ninja.BIN_DIR, 'ninja')}",
        ], cwd=build_temp, env=build_env)

        subprocess.check_call([
            "cmake",
            "--build",
            ".",
            "--target",
            "chromobius_pybind",
        ], cwd=build_temp, env=build_env)


__version__ = '1.2.dev1774659305'

with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

# HACK: Workaround difficulties collecting data files for the package by just making a directory.
package_data_dir = pathlib.Path(__file__).parent / 'package_data'
package_data_dir.mkdir(exist_ok=True)
doc_chromobius = pathlib.Path(__file__).parent / 'doc' / 'chromobius.pyi'
if doc_chromobius.exists():
    shutil.copyfile(
        pathlib.Path(__file__).parent / 'doc' / 'chromobius.pyi',
        package_data_dir / 'chromobius.pyi',
    )

setuptools.setup(
    name="chromobius",
    version=__version__,
    author="Craig Gidney",
    url="https://github.com/quantumlib/chromobius",
    description="A fast implementation of the Möbius color code decoder.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    maintainer="Google Quantum AI",
    maintainer_email="quantum-oss-maintainers@google.com",
    license="Apache-2.0",
    ext_modules=[CMakeExtension("chromobius", sourcedir=".", sources=RELEVANT_SOURCE_FILES)],
    cmdclass={"build_ext": CMakeBuild},
    python_requires=">=3.10",
    setup_requires=['ninja', 'pybind11~=2.11.1', 'cmake>=3.13'],
    install_requires=['numpy', 'stim'],

    # Needed on Windows to avoid the default `build` colliding with Bazel's `BUILD`.
    # Also, the replacement name is short to avoid blowing the 256 character path limit on windows.
    options={'build': {'build_base': 'b'}},

    # Add files in package_data_dir to the wheel.
    # I don't know why it has to be so esoteric, but I tried for hours.
    # This is the best I could come up with.
    packages=['chromobius'],
    package_dir={'chromobius': package_data_dir.name},
    package_data={'chromobius': [str(e) for e in package_data_dir.iterdir()]},
    include_package_data=True,
    classifiers=[
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Topic :: Scientific/Engineering :: Quantum Computing",
    ],
    keywords=[
        "algorithms",
        "color code",
        "fault-tolerant quantum computing",
        "möbius decoder",
        "quantum computing",
        "quantum error correction",
        "quantum",
    ],
)
